@@ -41,6 +41,11 @@ except ImportError:
             'install paths may be incorrect', file=sys.stderr)
 
 windows_build = os.name == 'nt'
+if windows_build:
+    import cx_Freeze
+    command_base = cx_Freeze
+else:
+    command_base = distutils
 
 
 def has_help(self):
@@ -59,7 +64,7 @@ def has_data(self):
     return "build_data" in self.distribution.cmdclass
 
 
-distutils.command.build.build.sub_commands.extend([
+command_base.command.build.Build.sub_commands.extend([
     ("build_i18n", has_i18n),
     ("build_icons", has_icons),
     ("build_help", has_help),
@@ -89,6 +94,8 @@ class build_data(distutils.cmd.Command):
         ('share/meld', ['data/gschemas.compiled']),
     ]
 
+    win32_settings_ini = '[Settings]\ngtk-application-prefer-dark-theme=0\n'
+
     style_source = "data/styles/*.style-scheme.xml.in"
     style_target_dir = 'share/meld/styles'
 
@@ -117,13 +124,25 @@ class build_data(distutils.cmd.Command):
             "glib-compile-resources",
             "--target={}".format(target),
             "--sourcedir={}".format(resource_dir),
+            "--sourcedir={}".format("data/icons/hicolor"),
             self.resource_source,
         ])
 
         data_files.append(('share/meld', [target]))
 
         if windows_build:
-            gschemas = self.frozen_gschemas
+            # Write out a default settings.ini for Windows to make
+            # e.g., dark theme selection slightly easier.
+            settings_dir = os.path.join('build', 'etc', 'gtk-3.0')
+            if not os.path.exists(settings_dir):
+                os.makedirs(settings_dir)
+            settings_path = os.path.join(settings_dir, 'settings.ini')
+            with open(settings_path, 'w') as f:
+                print(self.win32_settings_ini, file=f)
+
+            gschemas = self.frozen_gschemas + [
+                ('etc/gtk-3.0', [settings_path])
+            ]
         else:
             gschemas = self.gschemas
         data_files.extend(gschemas)
